#!/usr/bin/env python

from typing import Optional

import json

from hdwallet import HDWallet
from hdwallet.derivations import (
    Derivation, BIP32Derivation
)
from hdwallet.cryptocurrencies import (
    Cryptocurrency, get_cryptocurrency
)
from hdwallet.utils import generate_mnemonic
from hdwallet.cli import (
    click, sys
)


def generate_hdwallet(
    symbol: str,
    strength: Optional[int],
    entropy: Optional[str],
    mnemonic: Optional[str],
    language: Optional[str],
    passphrase: Optional[str],
    seed: Optional[str],
    xprivate_key: Optional[str],
    xpublic_key: Optional[str],
    strict: Optional[bool],
    account: int,
    change: bool,
    address: int,
    path: Optional[str],
    private_key: Optional[str],
    public_key: Optional[str],
    wif: Optional[str],
    semantic: str
):
    try:
        hdwallet: HDWallet = HDWallet(
            symbol=symbol, semantic=semantic
        )
        if entropy:
            hdwallet.from_entropy(
                entropy=entropy, language=language, passphrase=passphrase
            )
        elif mnemonic:
            hdwallet.from_mnemonic(
                mnemonic=mnemonic, language=language, passphrase=passphrase
            )
        elif seed:
            hdwallet.from_seed(
                seed=seed
            )
        elif xprivate_key:
            hdwallet.from_xprivate_key(
                xprivate_key=xprivate_key, strict=strict
            )
        elif xpublic_key:
            hdwallet.from_xpublic_key(
                xpublic_key=xpublic_key, strict=strict
            )
        elif private_key:
            hdwallet.from_private_key(
                private_key=private_key
            )
        elif public_key:
            hdwallet.from_public_key(
                public_key=public_key
            )
        elif wif:
            hdwallet.from_wif(
                wif=wif
            )
        else:
            mnemonic = generate_mnemonic(language=language, strength=strength)
            hdwallet.from_mnemonic(
                mnemonic=mnemonic, language=language, passphrase=passphrase
            )

        if wif or private_key or public_key:
            pass
        else:
            if path:
                derivation: Derivation = Derivation(path=path)
                hdwallet.from_path(path=derivation)
            else:
                cryptocurrency: Cryptocurrency = get_cryptocurrency(symbol=symbol)
                bip32_derivation: BIP32Derivation = BIP32Derivation(
                    purpose=(
                        44, False if xpublic_key else True
                    ),
                    coin_type=(
                        cryptocurrency.COIN_TYPE.INDEX,
                        False if xpublic_key else cryptocurrency.COIN_TYPE.HARDENED
                    ),
                    account=(
                        account, False if xpublic_key else True
                    ),
                    change=change,
                    address=address
                )
                hdwallet.from_path(path=bip32_derivation)

        click.echo(json.dumps(hdwallet.dumps(), indent=4, ensure_ascii=False))

    except Exception as exception:
        click.echo(click.style(f"Error: {str(exception)}"), err=True)
        sys.exit()



#!/usr/bin/env python3

import json
import os

from hdwallet.cli.__main__ import main as cli_main

# Test Values
base_path = os.path.dirname(__file__)
file_path = os.path.abspath(os.path.join(base_path, "..", "..", "values.json"))
values = open(file_path, "r", encoding="utf-8")
_ = json.loads(values.read())
values.close()


def test_addresses(cli_tester):

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate", "addresses",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--entropy", _["bitcoin"]["mainnet"]["entropy"],
            "--passphrase", _["bitcoin"]["mainnet"]["passphrase"],
            "--language", _["bitcoin"]["mainnet"]["language"]
        ]
    )

    dumps: dict = _["bitcoin"]["mainnet"]

    del dumps["root_xprivate_key_hex"]
    del dumps["root_xpublic_key_hex"]
    del dumps["xprivate_key_hex"]
    del dumps["xpublic_key_hex"]

    addresses: str = ""
    for address in _["bitcoin"]["addresses"]:
        addresses += f"{address}\n"

    assert hdwallet.exit_code == 0
    assert hdwallet.output == addresses

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate", "addresses",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--mnemonic", _["bitcoin"]["mainnet"]["mnemonic"],
            "--passphrase", _["bitcoin"]["mainnet"]["passphrase"],
            "--language", _["bitcoin"]["mainnet"]["language"]
        ]
    )

    assert hdwallet.exit_code == 0
    assert hdwallet.output == addresses

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate", "addresses",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--seed", _["bitcoin"]["mainnet"]["seed"]
        ]
    )

    assert hdwallet.exit_code == 0
    assert hdwallet.output == addresses

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate", "addresses",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--xprivate-key", _["bitcoin"]["mainnet"]["root_xprivate_key"]
        ]
    )

    assert hdwallet.exit_code == 0
    assert hdwallet.output == addresses

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate", "addresses",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--xpublic-key", _["bitcoin"]["mainnet"]["root_xpublic_key"]
        ]
    )

    assert hdwallet.exit_code == 0
    assert hdwallet.output



#!/usr/bin/env python

from tabulate import tabulate

from hdwallet.cli import click


def list_strengths():

    click.echo(tabulate(
        [
            [128, 12],
            [160, 15],
            [192, 18],
            [224, 21],
            [256, 24],
        ],
        [
            "Strength",
            "Words"
        ],
        tablefmt="github"
    ))



#!/usr/bin/env python3

from hdwallet.exceptions import (
    NetworkError, DerivationError, SemanticError, AddressError, SymbolError
)

import pytest


def test_exceptions():

    with pytest.raises(NetworkError, match="error"):
        raise NetworkError("error")
    with pytest.raises(NetworkError, match="error, error"):
        raise NetworkError("error", "error")
    with pytest.raises(DerivationError, match="error"):
        raise DerivationError("error")
    with pytest.raises(DerivationError, match="error, error"):
        raise DerivationError("error", "error")
    with pytest.raises(SemanticError, match="error"):
        raise SemanticError("error")
    with pytest.raises(SemanticError):
        raise SemanticError("error", "error")
    with pytest.raises(AddressError, match="error"):
        raise AddressError("error")
    with pytest.raises(AddressError, match="error, error"):
        raise AddressError("error", "error")
    with pytest.raises(SymbolError, match="error"):
        raise SymbolError("error")
    with pytest.raises(SymbolError, match="error, error"):
        raise SymbolError("error", "error")



#!/usr/bin/python3

from pathlib import Path
from click.testing import CliRunner

import os
import pytest


@pytest.fixture(scope="module")
def project_path():
    original_path = os.getcwd()
    os.chdir(original_path + "/tests")
    yield Path(original_path + "/tests")
    os.chdir(original_path)


@pytest.fixture(scope="module")
def cli_tester():
    return CliRunner()



#!/usr/bin/env python3

import json
import os

from hdwallet import HDWallet

# Test Values
base_path: str = os.path.dirname(__file__)
file_path: str = os.path.abspath(os.path.join(base_path, "../values.json"))
values = open(file_path, "r", encoding="utf-8")
_: dict = json.loads(values.read())
values.close()


def test_from_xpublic_key():

    hdwallet: HDWallet = HDWallet(
        symbol=_["bitcoin"]["mainnet"]["symbol"]
    )
    hdwallet.from_xpublic_key(
        xpublic_key=_["bitcoin"]["mainnet"]["root_xpublic_key"], strict=True
    )
    hdwallet.from_path(
        path="m/44/0/0/0/0"
    )

    assert hdwallet.cryptocurrency() == _["bitcoin"]["mainnet"]["cryptocurrency"]
    assert hdwallet.symbol() == _["bitcoin"]["mainnet"]["symbol"]
    assert hdwallet.network() == _["bitcoin"]["mainnet"]["network"]
    assert hdwallet.strength() is None
    assert hdwallet.entropy() is None
    assert hdwallet.mnemonic() is None
    assert hdwallet.language() is None
    assert hdwallet.passphrase() is None
    assert hdwallet.seed() is None
    assert hdwallet.root_xprivate_key(encoded=False) is None
    assert hdwallet.root_xprivate_key() is None
    assert hdwallet.root_xpublic_key(encoded=False) == "0488b21e000000000000000000ad41ef910cdcae932cb4060777b4284ee38f5b29c5fb60fda8416f298a14702c02949b9f64223e124eb9a8383fba0b21b5845fcfbdc84dec7692d21c716410eab0"
    assert hdwallet.root_xpublic_key() == "xpub661MyMwAqRbcGGUtsoFw2d6ARvD2ABd7z327zxt2XiBBwMx9GAuNrrE7tbRuWF5MjjZ1BzDsRdaSHc9nVKAgHzQrv6pwYW3Hd7LSzbh8sWS"
    assert hdwallet.xprivate_key(encoded=False) is None
    assert hdwallet.xprivate_key() is None
    assert hdwallet.xpublic_key(encoded=False) == "0488b21e052c0269af000000006c95c19e932b9e8f3d834e874526768ca1b3d89933ad71fd8253bcca67ac283d038f24175db513b40c75503c25040e5f0ea4d38e912d1f83daf5fd8c4b9512ad87"
    assert hdwallet.xpublic_key() == "xpub6FjoSaU1JaG6fC6wTYmb1mJzaZxSunxASN7nTRHhFynh33gKRfmmNrtQ82s8YouLCrEniskjumfACiiTyVmi4aXyLL8HvLdZc8mjKsbzT9z"
    assert hdwallet.uncompressed() == "8f24175db513b40c75503c25040e5f0ea4d38e912d1f83daf5fd8c4b9512ad8750a64d9e0ee3555225e4130c7e36a443ec20330bf0be1e4de913e31e00202993"
    assert hdwallet.compressed() == "038f24175db513b40c75503c25040e5f0ea4d38e912d1f83daf5fd8c4b9512ad87"
    assert hdwallet.chain_code() == "6c95c19e932b9e8f3d834e874526768ca1b3d89933ad71fd8253bcca67ac283d"
    assert hdwallet.private_key() is None
    assert hdwallet.public_key() == "038f24175db513b40c75503c25040e5f0ea4d38e912d1f83daf5fd8c4b9512ad87"
    assert hdwallet.wif() is None
    assert hdwallet.finger_print() == "4e749a26"
    assert hdwallet.semantic() == "p2pkh"
    assert hdwallet.path() == "m/44/0/0/0/0"
    assert hdwallet.hash() == "4e749a26934bca5091a05ee6f55e7d0e21482647"
    assert hdwallet.p2pkh_address() == "189qPd6J81ns9LEGx6kun7Xtg1bJV8GJXh"
    assert hdwallet.p2sh_address() == "3C71bNRojv3Gc7zHvWygas4AFt34rKezcF"
    assert hdwallet.p2wpkh_address() == "bc1qfe6f5f5nf099pydqtmn02hnapcs5sfj86dpqjm"
    assert hdwallet.p2wpkh_in_p2sh_address() == "3NykoodgJ7Li43JPt5xsezQz8xfwwwFZUs"
    assert hdwallet.p2wsh_address() == "bc1qazm6kznlgs06exh4cq2qxh567xrffppwujje5zg84upnng4essusd08nhz"
    assert hdwallet.p2wsh_in_p2sh_address() == "32yGj8ncXBBTjXqg188ZHxd1xffoQDcjin"

    assert isinstance(hdwallet.dumps(), dict)

    hdwallet: HDWallet = HDWallet(
        symbol=_["bitcoin"]["testnet"]["symbol"]
    )

    hdwallet.from_xpublic_key(
        xpublic_key=_["bitcoin"]["testnet"]["xpublic_key"], strict=False
    )

    assert hdwallet.cryptocurrency() == _["bitcoin"]["testnet"]["cryptocurrency"]
    assert hdwallet.symbol() == _["bitcoin"]["testnet"]["symbol"]
    assert hdwallet.network() == _["bitcoin"]["testnet"]["network"]
    assert hdwallet.strength() is None
    assert hdwallet.entropy() is None
    assert hdwallet.mnemonic() is None
    assert hdwallet.language() is None
    assert hdwallet.passphrase() is None
    assert hdwallet.seed() is None
    assert hdwallet.root_xprivate_key(encoded=False) is None
    assert hdwallet.root_xprivate_key() is None
    assert hdwallet.root_xpublic_key(encoded=False) == _["bitcoin"]["testnet"]["xpublic_key_hex"]
    assert hdwallet.root_xpublic_key() == _["bitcoin"]["testnet"]["xpublic_key"]
    assert hdwallet.xprivate_key(encoded=False) is None
    assert hdwallet.xprivate_key() is None
    assert hdwallet.xpublic_key(encoded=False) == _["bitcoin"]["testnet"]["xpublic_key_hex"]
    assert hdwallet.xpublic_key() == _["bitcoin"]["testnet"]["xpublic_key"]
    assert hdwallet.uncompressed() == _["bitcoin"]["testnet"]["uncompressed"]
    assert hdwallet.uncompressed(compressed=_["bitcoin"]["testnet"]["compressed"]) == _["bitcoin"]["testnet"]["uncompressed"]
    assert hdwallet.compressed() == _["bitcoin"]["testnet"]["compressed"]
    assert hdwallet.compressed(uncompressed=_["bitcoin"]["testnet"]["uncompressed"]) == _["bitcoin"]["testnet"]["compressed"]
    assert hdwallet.chain_code() == _["bitcoin"]["testnet"]["chain_code"]
    assert hdwallet.private_key() is None
    assert hdwallet.public_key() == _["bitcoin"]["testnet"]["public_key"]
    assert hdwallet.wif() is None
    assert hdwallet.finger_print() == _["bitcoin"]["testnet"]["finger_print"]
    assert hdwallet.semantic() == _["bitcoin"]["testnet"]["semantic"]
    assert hdwallet.path() == None
    assert hdwallet.hash() == _["bitcoin"]["testnet"]["hash"]
    assert hdwallet.p2pkh_address() == _["bitcoin"]["testnet"]["addresses"]["p2pkh"]
    assert hdwallet.p2sh_address() == _["bitcoin"]["testnet"]["addresses"]["p2sh"]
    assert hdwallet.p2wpkh_address() == _["bitcoin"]["testnet"]["addresses"]["p2wpkh"]
    assert hdwallet.p2wpkh_in_p2sh_address() == _["bitcoin"]["testnet"]["addresses"]["p2wpkh_in_p2sh"]
    assert hdwallet.p2wsh_address() == _["bitcoin"]["testnet"]["addresses"]["p2wsh"]
    assert hdwallet.p2wsh_in_p2sh_address() == _["bitcoin"]["testnet"]["addresses"]["p2wsh_in_p2sh"]

    assert isinstance(hdwallet.dumps(), dict)

    dumps: dict = _["bitcoin"]["testnet"]

    dumps["strength"] = None
    dumps["entropy"] = None
    dumps["mnemonic"] = None
    dumps["language"] = None
    dumps["passphrase"] = None
    dumps["seed"] = None
    dumps["root_xprivate_key"] = None
    dumps["xprivate_key"] = None
    dumps["private_key"] = None
    dumps["wif"] = None
    dumps["path"] = None
    dumps["root_xpublic_key"] = _["bitcoin"]["testnet"]["xpublic_key"]
    del dumps["root_xprivate_key_hex"]
    del dumps["root_xpublic_key_hex"]
    del dumps["xprivate_key_hex"]
    del dumps["xpublic_key_hex"]

    assert hdwallet.dumps() == dumps



# !/usr/bin/env python3

# Anon
ANON = "ANON"
# Argoneum
AGM = "AGM"
# Artax
XAX = "XAX"
# Aryacoin
AYA = "AYA"
# Asiacoin
AC = "AC"
# Atom
ATOM = "ATOM"
# Auroracoin
AUR = "AUR"
# Axe
AXE = "AXE"
# Bata
BTA = "BTA"
# Beetle Coin
BEET = "BEET"
# Bela Coin
BELA = "BELA"
# Bit Cloud
BTDX = "BTDX"
# Bit Send
BSD = "BSD"
# Bitcoin Cash
BCH = "BCH"
# Bitcoin Gold
BTG = "BTG"
# Bitcoin
BTC, BTCTEST = "BTC", "BTCTEST"
# Bitcoin Plus
XBC = "XBC"
# Bitcoin SV
BSV = "BSV"
# BitcoinZ
BTCZ = "BTCZ"
# Bitcore
BTX = "BTX"
# Blackcoin
BLK = "BLK"
# Block Stamp
BST = "BST"
# Blocknode
BND, BNDTEST = "BND", "BNDTEST"
# Bolivarcoin
BOLI = "BOLI"
# Brit Coin
BRIT = "BRIT"
# CPU Chain
CPU = "CPU"
# Canada eCoin
CDN = "CDN"
# Cannacoin
CCN = "CCN"
# Clams
CLAM = "CLAM"
# Club Coin
CLUB = "CLUB"
# Compcoin
CMP = "CMP"
# Crane Pay
CRP = "CRP"
# Crave
CRAVE = "CRAVE"
# Dash
DASH, DASHTEST = "DASH", "DASHTEST"
# Deep Onion
ONION = "ONION"
# Defcoin
DFC = "DFC"
# Denarius
DNR = "DNR"
# Diamond
DMD = "DMD"
# Digi Byte
DGB = "DGB"
# Digitalcoin
DGC = "DGC"
# Dogecoin
DOGE, DOGETEST = "DOGE", "DOGETEST"
# EDR Coin
EDRC = "EDRC"
# Ecoin
ECN = "ECN"
# Einsteinium
EMC2 = "EMC2"
# Elastos
ELA = "ELA"
# Energi
NRG = "NRG"
# Ethereum
ETH = "ETH"
# Europe Coin
ERC = "ERC"
# Exclusive Coin
EXCL = "EXCL"
# FIX
FIX, FIXTEST = "FIX", "FIXTEST"
# Feathercoin
FTC = "FTC"
# Firstcoin
FRST = "FRST"
# Flashcoin
FLASH = "FLASH"
# Fuji Coin
FJC = "FJC"
# GCR Coin
GCR = "GCR"
# Game Credits
GAME = "GAME"
# Go Byte
GBX = "GBX"
# Gridcoin
GRC = "GRC"
# Groestl Coin
GRS, GRSTEST = "GRS", "GRSTEST"
# Gulden
NLG = "NLG"
# Helleniccoin
HNC = "HNC"
# Hempcoin
THC = "THC"
# Hush
HUSH = "HUSH"
# IX Coin
IXC = "IXC"
# Insane Coin
INSN = "INSN"
# Internet Of People
IOP = "IOP"
# Jumbucks
JBS = "JBS"
# Kobocoin
KOBO = "KOBO"
# Komodo
KMD = "KMD"
# LBRY Credits
LBC = "LBC"
# Linx
LINX = "LINX"
# Litecoin Cash
LCC = "LCC"
# Litecoin
LTC, LTCTEST = "LTC", "LTCTEST"
# LitecoinZ
LTZ = "LTZ"
# Lkrcoin
LKR = "LKR"
# Lynx
LYNX = "LYNX"
# Mazacoin
MZC = "MZC"
# Megacoin
MEC = "MEC"
# Minexcoin
MNX = "MNX"
# Monacoin
MONA = "MONA"
# Monkey Project
MONK = "MONK"
# Myriadcoin
XMY = "XMY"
# NIX
NIX = "NIX"
# Namecoin
NMC = "NMC"
# Navcoin
NAV = "NAV"
# Neblio
NEBL = "NEBL"
# Neoscoin
NEOS = "NEOS"
# Neurocoin
NRO = "NRO"
# New York Coin
NYC = "NYC"
# Novacoin
NVC = "NVC"
# NuBits
NBT = "NBT"
# NuShares
NSR = "NSR"
# OK Cash
OK = "OK"
# Omni
OMNI, OMNITEST = "OMNI", "OMNITEST"
# Onix Coin
ONX = "ONX"
# Peercoin
PPC = "PPC"
# Pesobit
PSB = "PSB"
# Phore
PHR = "PHR"
# Pinkcoin
PINK = "PINK"
# Pivx
PIVX, PIVXTEST = "PIVX", "PIVXTEST"
# Posw Coin
POSW = "POSW"
# Potcoin
POT = "POT"
# Project Coin
PRJ = "PRJ"
# Putincoin
PUT = "PUT"
# Qtum
QTUM, QTUMTEST = "QTUM", "QTUMTEST"
# RSK
RBTC, RBTCTEST = "RBTC", "RBTCTEST"
# Rapids
RPD = "RPD"
# Ravencoin
RVN = "RVN"
# Reddcoin
RDD = "RDD"
# Rubycoin
RBY = "RBY"
# Safecoin
SAFE = "SAFE"
# Saluscoin
SLS = "SLS"
# Scribe
SCRIBE = "SCRIBE"
# Shadow Cash
SDC, SDCTEST = "SDC", "SDCTEST"
# Slimcoin
SLM, SLMTEST = "SLM", "SLMTEST"
# Smileycoin
SMLY = "SMLY"
# Solarcoin
SLR = "SLR"
# Stash
STASH = "STASH"
# Stratis
STRAT, STRATTEST = "STRAT", "STRATTEST"
# Sugarchain
SUGAR, SUGARTEST = "SUGAR", "SUGARTEST"
# Syscoin
SYS = "SYS"
# TOA Coin
TOA = "TOA"
# Thought AI
THT = "THT"
# TRX
TRX = "TRX"
# Twins
TWINS, TWINSTEST = "TWINS", "TWINSTEST"
# Ultimate Secure Cash
USC = "USC"
# Unobtanium
UNO = "UNO"
# Virtual Cash
VASH = "VASH"
# Vcash
VC = "VC"
# Verge Currency
XVG = "XVG"
# Vertcoin
VTC = "VTC"
# Viacoin
VIA, VIATEST = "VIA", "VIATEST"
# Vivo
VIVO = "VIVO"
# Whitecoin
XWC = "XWC"
# Wincoin
WC = "WC"
# XUEZ
XUEZ = "XUEZ"
# XinFin
XDC = "XDC"
# ZClassic
ZCL = "ZCL"
# Zcash
ZEC, ZECTEST = "ZEC", "ZECTEST"
# Zencash
ZEN = "ZEN"


__all__ = [
    "ANON",
    "AGM",
    "XAX",
    "AYA",
    "AC",
    "ATOM",
    "AUR",
    "AXE",
    "BTA",
    "BEET",
    "BELA",
    "BTDX",
    "BSD",
    "BCH",
    "BTG",
    "BTC", "BTCTEST",
    "XBC",
    "BSV",
    "BTCZ",
    "BTX",
    "BLK",
    "BST",
    "BND", "BNDTEST",
    "BOLI",
    "BRIT",
    "CPU",
    "CDN",
    "CCN",
    "CLAM",
    "CLUB",
    "CMP",
    "CRP",
    "CRAVE",
    "DASH", "DASHTEST",
    "ONION",
    "DFC",
    "DNR",
    "DMD",
    "DGB",
    "DGC",
    "DOGE", "DOGETEST",
    "EDRC",
    "ECN",
    "EMC2",
    "ELA",
    "NRG",
    "ETH",
    "ERC",
    "EXCL",
    "FIX", "FIXTEST",
    "FTC",
    "FRST",
    "FLASH",
    "FJC",
    "GCR",
    "GAME",
    "GBX",
    "GRC",
    "GRS", "GRSTEST",
    "NLG",
    "HNC",
    "THC",
    "HUSH",
    "IXC",
    "INSN",
    "IOP",
    "JBS",
    "KOBO",
    "KMD",
    "LBC",
    "LINX",
    "LCC",
    "LTC", "LTCTEST",
    "LTZ",
    "LKR",
    "LYNX",
    "MZC",
    "MEC",
    "MNX",
    "MONA",
    "MONK",
    "XMY",
    "NIX",
    "NMC",
    "NAV",
    "NEBL",
    "NEOS",
    "NRO",
    "NYC",
    "NVC",
    "NBT",
    "NSR",
    "OK",
    "OMNI", "OMNITEST",
    "ONX",
    "PPC",
    "PSB",
    "PHR",
    "PINK",
    "PIVX", "PIVXTEST",
    "POSW",
    "POT",
    "PRJ",
    "PUT",
    "QTUM", "QTUMTEST",
    "RBTC", "RBTCTEST",
    "RPD",
    "RVN",
    "RDD",
    "RBY",
    "SAFE",
    "SLS",
    "SCRIBE",
    "SDC", "SDCTEST",
    "SLM", "SLMTEST",
    "SMLY",
    "SLR",
    "STASH",
    "STRAT", "STRATTEST",
    "SUGAR", "SUGARTEST",
    "SYS",
    "TOA",
    "THT",
    "TRX",
    "TWINS", "TWINSTEST",
    "USC",
    "UNO",
    "VASH",
    "VC",
    "XVG",
    "VTC",
    "VIA", "VIATEST",
    "VIVO",
    "XWC",
    "WC",
    "XUEZ",
    "XDC",
    "ZCL",
    "ZEC", "ZECTEST",
    "ZEN"
]



#!/usr/bin/env python3

from typing import (
    Union, Tuple, Any, Optional
)

from .cryptocurrencies import Cryptocurrency
from .exceptions import (
    SemanticError, DerivationError
)

HARDENED: Tuple[str, str] = ("'", "")


class Derivation:
    """
    Hierarchical Deterministic Wallet Derivation's

    :param path: Derivation path.
    :type path: str
    :param semantic: Extended semantic, defaults to ``P2PKH``.
    :type semantic: str

    :returns: Derivation -- Derivation instance.

    >>> from hdwallet.derivations import Derivation
    >>> Derivation()
    <hdwallet.derivations.Derivation object at 0x000001EBC58E9F70>
    >>> str(Derivation())
    "\0\0\0\0"
    >>> str(Derivation(path="m/44'/0'/0'/0/0", semantic="p2pkh"))
    "m/44'/0'/0'/0/0"

    .. note::
        Do not forget all derivation paths are start swith 'm/' prefix.
    """

    PATH: str = "\0\0\0\0"
    SEMANTIC: str = "p2pkh"

    def __str__(self) -> str:
        return self.PATH

    def __init__(self, path: Optional[str] = None, semantic: str = "p2pkh"):

        if path:
            if not isinstance(path, str):
                raise DerivationError("Bad derivation path, Please import only str type!")
            elif path[0:2] != "m/":
                raise DerivationError("Bad path, please insert like this str type of \"m/0'/0\" path!")

            self.PATH = "m"
            for index in path.lstrip("m/").split("/"):
                self.PATH += f"/{int(index[:-1])}'" if "'" in index else f"/{int(index)}"

        self.SEMANTIC = semantic

    @classmethod
    def from_path(cls, path: str) -> "Derivation":
        """
        Derivation from path.

        :param path: Derivation path.
        :type path: str, Derivation

        :returns: Derivation -- Derivation instance.

        >>> from hdwallet.derivations import Derivation
        >>> derivation = Derivation()
        >>> derivation.from_path(path="m/44'/0'/'0/0/0")
        <hdwallet.derivation.Derivation object at 0x000001E8BFB98D60>
        """

        if not isinstance(path, str):
            raise DerivationError("Bad derivation path, Please import only str type!")
        if path[0:2] != "m/":
            raise DerivationError("Bad path, please insert like this str type of \"m/0'/0\" path!")

        new_path = "m"
        for index in path.lstrip("m/").split("/"):
            new_path += f"/{int(index[:-1])}'" if "'" in index else f"/{int(index)}"

        return Derivation(path=new_path)

    def from_index(self, index: int, hardened: bool = False) -> "Derivation":
        """
        Derivation from path.

        :param index: Derivation index.
        :type index: int
        :param hardened: Hardened address, default to ``False``.
        :type hardened: bool

        :returns: Derivation -- Derivation instance.

        >>> from hdwallet.derivations import Derivation
        >>> derivation = Derivation()
        >>> derivation.from_index(index=44, hardened=True)
        >>> derivation.from_index(index=0, hardened=True)
        >>> derivation.from_index(index=0, hardened=True)
        >>> derivation.from_index(index=0)
        >>> derivation.from_index(index=0)
        <hdwallet.derivation.Derivation object at 0x000001E8BFB98D60>
        """

        if not isinstance(index, int):
            raise DerivationError("Bad derivation index, Please import only int type!")

        if self.PATH == "\0\0\0\0":
            self.PATH = ""
        self.PATH += (
            (f"/{index}'" if hardened else f"/{index}")
            if self.PATH.startswith("m/") else
            (f"m/{index}'" if hardened else f"m/{index}")
        )
        return self

    def clean_derivation(self) -> "Derivation":
        """
        Clean derivation path or indexes.

        :returns: Derivation -- Derivation instance.

        >>> from hdwallet.derivations import Derivation
        >>> derivation = Derivation()
        >>> derivation.from_path(path="m/44'/0'/'0/0/0")
        >>> str(derivation)
        "m/44'/0'/'0/0/0"
        >>> derivation.clean_derivation()
        <hdwallet.wallet.HDWallet object at 0x000001E8BFB98D60>
        >>> str(derivation)
        "\0\0\0\0"
        """

        self.PATH = "\0\0\0\0"
        return self


class BIP32Derivation(Derivation):
    """
    Hierarchical Deterministic Wallet BIP32 Derivation

    :param cryptocurrency: Cryptocurrency instance, default to ``None``.
    :type cryptocurrency: Cryptocurrency
    :param purpose: Purpose index, default to ``0``.
    :type purpose: int, tuple
    :param coin_type: Coin type, default to ``0``.
    :type coin_type: int, tuple
    :param account: Account index, default to ``0``.
    :type account: int, tuple
    :param change: Change addresses, default to ``False``.
    :type change: bool
    :param address: Address index, default to ``0``.
    :type address: int, tuple

    :returns: BIP32Derivation -- BIP32Derivation instance.

    >>> from hdwallet.derivations import BIP32Derivation
    >>> from hdwallet.cryptocurrencies import BitcoinMainnet
    >>> BIP32Derivation(cryptocurrency=BitcoinMainnet)
    <hdwallet.derivations.Derivation object at 0x000001EBC58E9F70>
    >>> str(BIP32Derivation(cryptocurrency=BitcoinMainnet))
    "m/0'/0'/0'/0/0"
    """

    PURPOSE: Tuple[int, bool]
    COIN_TYPE: Tuple[int, bool]
    ACCOUNT: Tuple[int, bool]
    CHANGE: bool
    ADDRESS: Tuple[int, bool]

    def __str__(self):
        if self.PATH == "\0\0\0\0":
            return f"m/{self.PURPOSE[0]}{HARDENED[0] if self.PURPOSE[1] else HARDENED[1]}" \
                   f"/{self.COIN_TYPE[0]}{HARDENED[0] if self.COIN_TYPE[1] else HARDENED[1]}" \
                   f"/{self.ACCOUNT[0]}{HARDENED[0] if self.ACCOUNT[1] else HARDENED[1]}" \
                   f"/{1 if self.CHANGE else 0}" \
                   f"/{self.ADDRESS[0]}{HARDENED[0] if self.ADDRESS[1] else HARDENED[1]}"
        return self.PATH

    def __init__(self, cryptocurrency: Any = None,
                 purpose: Union[int, Tuple[int, bool]] = 0,
                 coin_type: Union[int, Tuple[int, bool]] = 0,
                 account: Union[int, Tuple[int, bool]] = 0,
                 change: bool = False,
                 address: Union[int, Tuple[int, bool]] = 0):

        super(BIP32Derivation, self).__init__()

        self.PURPOSE, self.COIN_TYPE, self.ACCOUNT, self.CHANGE, self.ADDRESS = (
            purpose if isinstance(purpose, tuple) else (purpose, True),
            (
                (cryptocurrency.COIN_TYPE.INDEX, cryptocurrency.COIN_TYPE.HARDENED)
                if cryptocurrency else
                (coin_type if isinstance(coin_type, tuple) else (coin_type, True))
            ),
            account if isinstance(account, tuple) else (account, True),
            change,
            address if isinstance(address, tuple) else (address, False)
        )

        self.SEMANTIC = "p2pkh"

    def from_purpose(self, purpose: int, hardened: bool = True) -> "BIP32Derivation":
        """
        Derivation from purpose index.

        :param purpose: Purpose index.
        :type purpose: int
        :param hardened: Hardened, default to ``True``.
        :type hardened: bool

        :returns: BIP32Derivation -- BIP32Derivation instance.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_purpose(purpose=0, hardened=True)
        <hdwallet.derivation.BIP32Derivation object at 0x000001E8BFB98D60>
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("When you are using custom path, you can't set purpose.")
        if type(self).__name__ != "BIP32Derivation":
            raise TypeError(
                f"You can't set purpose for {type(self).__name__}, it's only for BIP32Derivation class."
            )
        self.PURPOSE = (purpose, hardened)
        return self

    def from_coin_type(self, coin_type: int, hardened: bool = True) -> "BIP32Derivation":
        """
        Derivation from Coin Type index.

        :param coin_type: Coin type index.
        :type coin_type: int
        :param hardened: Hardened, default to ``True``.
        :type hardened: bool

        :returns: BIP32Derivation -- BIP32Derivation instance.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_coin_type(coin_type=56, hardened=True)
        <hdwallet.derivation.BIP32Derivation object at 0x000001E8BFB98D60>
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("When you are using custom path, you can't set coin type.")
        self.COIN_TYPE = (coin_type, hardened)
        return self

    def from_account(self, account: int, hardened: bool = True) -> "BIP32Derivation":
        """
        Derivation from Account index.

        :param account: Coin type index.
        :type account: int
        :param hardened: Hardened, default to ``True``.
        :type hardened: bool

        :returns: BIP32Derivation -- BIP32Derivation instance.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_account(account=1, hardened=True)
        <hdwallet.derivation.BIP32Derivation object at 0x000001E8BFB98D60>
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("When you are using custom path, you can't set account.")
        self.ACCOUNT = (account, hardened)
        return self

    def from_change(self, change: bool) -> "BIP32Derivation":
        """
        Derivation from external Change.

        :param change: External chnage.
        :type change: bool

        :returns: BIP32Derivation -- BIP32Derivation instance.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_account(change=True)
        <hdwallet.derivation.BIP32Derivation object at 0x000001E8BFB98D60>
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("When you are using custom path, you can't set change.")
        self.CHANGE = change
        return self

    def from_address(self, address: int, hardened: bool = False) -> "BIP32Derivation":
        """
        Derivation from Address index.

        :param address: Address index.
        :type address: int
        :param hardened: Hardened, default to ``True``.
        :type hardened: bool

        :returns: BIP32Derivation -- BIP32Derivation instance.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_address(address=1, hardened=True)
        <hdwallet.derivation.BIP32Derivation object at 0x000001E8BFB98D60>
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("When you are using custom path, you can't set address.")
        self.ADDRESS = (address, hardened)
        return self

    def clean_derivation(self) -> "BIP32Derivation":
        self.PURPOSE, self.COIN_TYPE, self.ACCOUNT, self.CHANGE, self.ADDRESS = (
            (0, True), (0, True), (0, True), False, (0, False)
        )
        self.PATH = "\0\0\0\0"
        return self

    def purpose(self) -> str:
        """
        Gey Purpose index.

        :returns: str -- Purpose index.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_purpose(purpose=0, hardened=True)
        >>> bip32_derivation.purpose()
        "0'"
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("You can't get purpose from custom path.")
        return f"{self.PURPOSE[0]}{HARDENED[0] if self.PURPOSE[1] else HARDENED[1]}"

    def coin_type(self) -> str:
        """
        Gey Coin Type index.

        :returns: str -- Coin Type index.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.coin_type(coin_type=15, hardened=True)
        >>> bip32_derivation.coin_type()
        "15'"
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("You can't get coin type from custom path.")
        return f"{self.COIN_TYPE[0]}{HARDENED[0] if self.COIN_TYPE[1] else HARDENED[1]}"

    def account(self) -> str:
        """
        Gey Account index.

        :returns: str -- Account index.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_account(account=1, hardened=True)
        >>> bip32_derivation.account()
        "1'"
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("You can't get account from custom path.")
        return f"{self.ACCOUNT[0]}{HARDENED[0] if self.ACCOUNT[1] else HARDENED[1]}"

    def change(self, number: bool = False) -> Union[str, bool]:
        """
        Gey external Change.

        :param number: Return type, default to ``False``.
        :type number: bool

        :returns: str -- External change index.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_change(change=True)
        >>> bip32_derivation.change(number=True)
        "1"
        >>> bip32_derivation.change(number=False)
        True
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("You can't get change from custom path.")
        return f"{1 if self.CHANGE else 0}" if number else self.CHANGE

    def address(self) -> str:
        """
        Gey Address index.

        :returns: str -- Address index.

        >>> from hdwallet.derivations import BIP32Derivation
        >>> bip32_derivation = BIP32Derivation()
        >>> bip32_derivation.from_address(address=1, hardened=True)
        >>> bip32_derivation.address()
        "1"
        """

        if self.PATH and self.PATH != "\0\0\0\0":
            raise DerivationError("You can't get address from custom path.")
        return f"{self.ADDRESS[0]}{HARDENED[0] if self.ADDRESS[1] else HARDENED[1]}"


class BIP44Derivation(BIP32Derivation):
    """
    Hierarchical Deterministic Wallet BIP44 Derivation

    :param cryptocurrency: Cryptocurrency instance.
    :type cryptocurrency: Cryptocurrency
    :param account: Account index, default to ``0``.
    :type account: int, tuple
    :param change: Change addresses, default to ``False``.
    :type change: bool
    :param address: Address index, default to ``0``.
    :type address: int, tuple

    :returns: BIP44Derivation -- BIP44Derivation instance.

    >>> from hdwallet.derivations import BIP44Derivation
    >>> from hdwallet.cryptocurrencies import BitcoinMainnet
    >>> BIP44Derivation(cryptocurrency=BitcoinMainnet)
    <hdwallet.derivations.Derivation object at 0x000001EBC58E9F70>
    >>> str(BIP44Derivation(cryptocurrency=BitcoinMainnet))
    "m/44'/0'/0'/0/0"
    """

    PURPOSE: int = 44

    def __init__(self, cryptocurrency: Any,
                 account: Union[int, Tuple[int, bool]] = 0,
                 change: bool = False,
                 address: Union[int, Tuple[int, bool]] = 0):
        super().__init__(
            cryptocurrency=cryptocurrency,
            purpose=self.PURPOSE,
            coin_type=cryptocurrency.COIN_TYPE,
            account=account,
            change=change,
            address=address
        )

        self.SEMANTIC = "p2pkh"


class BIP49Derivation(BIP32Derivation):
    """
    Hierarchical Deterministic Wallet BIP49 Derivation

    :param cryptocurrency: Cryptocurrency instance.
    :type cryptocurrency: Cryptocurrency
    :param account: Account index, default to ``0``.
    :type account: int, tuple
    :param change: Change addresses, default to ``False``.
    :type change: bool
    :param address: Address index, default to ``0``.
    :type address: int, tuple

    :returns: BIP49Derivation -- BIP49Derivation instance.

    >>> from hdwallet.derivations import BIP49Derivation
    >>> from hdwallet.cryptocurrencies import BitcoinMainnet
    >>> BIP49Derivation(cryptocurrency=BitcoinMainnet)
    <hdwallet.derivations.Derivation object at 0x000001EBC58E9F70>
    >>> str(BIP49Derivation(cryptocurrency=BitcoinMainnet))
    "m/49'/0'/0'/0/0"
    """

    PURPOSE: int = 49

    def __init__(self, cryptocurrency: Any,
                 account: Union[int, Tuple[int, bool]] = 0,
                 change: bool = False,
                 address: Union[int, Tuple[int, bool]] = 0):
        super(BIP49Derivation, self).__init__(
            cryptocurrency=cryptocurrency,
            purpose=self.PURPOSE,
            account=account,
            change=change,
            address=address
        )

        self.SEMANTIC = "p2wpkh_in_p2sh"


class BIP84Derivation(BIP32Derivation):
    """
    Hierarchical Deterministic Wallet BIP84 Derivation

    :param cryptocurrency: Cryptocurrency instance.
    :type cryptocurrency: Cryptocurrency
    :param account: Account index, default to ``0``.
    :type account: int, tuple
    :param change: Change addresses, default to ``False``.
    :type change: bool
    :param address: Address index, default to ``0``.
    :type address: int, tuple

    :returns: BIP84Derivation -- BIP84Derivation instance.

    >>> from hdwallet.derivations import BIP84Derivation
    >>> from hdwallet.cryptocurrencies import BitcoinMainnet
    >>> BIP84Derivation(cryptocurrency=BitcoinMainnet)
    <hdwallet.derivations.Derivation object at 0x000001EBC58E9F70>
    >>> str(BIP84Derivation(cryptocurrency=BitcoinMainnet))
    "m/84'/0'/0'/0/0"
    """

    PURPOSE: int = 84

    def __init__(self, cryptocurrency: Any,
                 account: Union[int, Tuple[int, bool]] = 0,
                 change: bool = False,
                 address: Union[int, Tuple[int, bool]] = 0):
        super().__init__(
            cryptocurrency=cryptocurrency,
            purpose=self.PURPOSE,
            coin_type=cryptocurrency.COIN_TYPE,
            account=account,
            change=change,
            address=address
        )

        self.SEMANTIC = "p2wpkh"


class BIP141Derivation(Derivation):
    """
    Hierarchical Deterministic Wallet BIP141 Derivation

    :param path: Derivation path, default to ``None``.
    :type path: str
    :param semantic: Extended semantic, defaults to ``P2WPKH``.
    :type semantic: str

    :returns: BIP141Derivation -- BIP141Derivation instance.

    >>> from hdwallet.derivations import BIP141Derivation
    >>> from hdwallet.cryptocurrencies import BitcoinMainnet
    >>> BIP141Derivation(cryptocurrency=BitcoinMainnet)
    <hdwallet.derivations.Derivation object at 0x000001EBC58E9F70>
    >>> str(BIP141Derivation(cryptocurrency=BitcoinMainnet))
    "m/44'/0'/0'/0/0"
    """

    def __init__(self, cryptocurrency: Any, path: Union[str, Derivation] = None, semantic: str = "p2wpkh"):

        if semantic not in ["p2wpkh", "p2wpkh_in_p2sh", "p2wsh", "p2wsh_in_p2sh"]:
            raise SemanticError(
                "Wrong extended semantic",
                "choose only the following options 'p2wpkh', 'p2wpkh_in_p2sh', 'p2wsh' or 'p2wsh_in_p2sh' semantics."
            )

        super(BIP141Derivation, self).__init__(
            path=(path if path else cryptocurrency.DEFAULT_PATH), semantic=semantic
        )
        self.SEMANTIC = semantic



from io import BytesIO

import hashlib
import hmac

A = 0
B = 7
P = 2**256 - 2**32 - 977
N = 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141
BASE58_ALPHABET = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def encode_base58(data):
    """
    This function encodes bytes to a base58 string
    """
    # determine how many 0 bytes (b'\x00') s starts with
    count = 0
    for byte in data:
        if byte == 0:
            count += 1
        else:
            break
    # convert to big endian integer
    num = int.from_bytes(data, 'big')
    prefix = '1' * count
    result = ''
    while num > 0:
        num, mod = divmod(num, 58)
        result = BASE58_ALPHABET[mod] + result
    return prefix + result


def encode_base58_checksum(data):
    """
    This function returns the Base58 check format
    """
    return encode_base58(data + hash256(data)[:4])


def hash160(data):
    """sha256 followed by ripemd160"""
    return hashlib.new('ripemd160', hashlib.sha256(data).digest()).digest()


def hash256(data):
    """two rounds of sha256"""
    return hashlib.sha256(hashlib.sha256(data).digest()).digest()


def sha256(data):
    """one round of sha256"""
    return hashlib.sha256(data).digest()


class FieldElement:

    def __init__(self, num, prime):
        if num >= prime or num < 0:
            error = 'Num {} not in field range 0 to {}'.format(
                num, prime - 1)
            raise ValueError(error)
        self.num = num
        self.prime = prime

    def __repr__(self):
        return 'FieldElement_{}({})'.format(self.prime, self.num)

    def __eq__(self, other):
        if other is None:
            return False
        return self.num == other.num and self.prime == other.prime

    def __ne__(self, other):
        # this should be the inverse of the == operator
        return not (self == other)

    def __add__(self, other):
        if self.prime != other.prime:
            raise TypeError('Cannot add two numbers in different Fields')
        # self.num and other.num are the actual values
        # self.prime is what we need to mod against
        num = (self.num + other.num) % self.prime
        # We return an element of the same class
        return self.__class__(num, self.prime)

    def __sub__(self, other):
        if self.prime != other.prime:
            raise TypeError('Cannot subtract two numbers in different Fields')
        # self.num and other.num are the actual values
        # self.prime is what we need to mod against
        num = (self.num - other.num) % self.prime
        # We return an element of the same class
        return self.__class__(num, self.prime)

    def __mul__(self, other):
        if self.prime != other.prime:
            raise TypeError('Cannot multiply two numbers in different Fields')
        # self.num and other.num are the actual values
        # self.prime is what we need to mod against
        num = (self.num * other.num) % self.prime
        # We return an element of the same class
        return self.__class__(num, self.prime)

    def __pow__(self, exponent):
        n = exponent % (self.prime - 1)
        num = pow(self.num, n, self.prime)
        return self.__class__(num, self.prime)

    def __truediv__(self, other):
        if self.prime != other.prime:
            raise TypeError('Cannot divide two numbers in different Fields')
        # self.num and other.num are the actual values
        # self.prime is what we need to mod against
        # use fermat's little theorem:
        # self.num**(p-1) % p == 1
        # this means:
        # 1/n == pow(n, p-2, p)
        num = (self.num * pow(other.num, self.prime - 2, self.prime)) % self.prime
        # We return an element of the same class
        return self.__class__(num, self.prime)

    def __rmul__(self, coefficient):
        num = (self.num * coefficient) % self.prime
        return self.__class__(num=num, prime=self.prime)


class Point:

    def __init__(self, x, y, a, b):
        self.a = a
        self.b = b
        self.x = x
        self.y = y
        # x being None and y being None represents the point at infinity
        # Check for that here since the equation below won't make sense
        # with None values for both.
        if self.x is None and self.y is None:
            return
        # make sure that the elliptic curve equation is satisfied
        # y**2 == x**3 + a*x + b
        if self.y**2 != self.x**3 + a * x + b:
            # if not, throw a ValueError
            raise ValueError('({}, {}) is not on the curve'.format(x, y))

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y \
            and self.a == other.a and self.b == other.b

    def __ne__(self, other):
        # this should be the inverse of the == operator
        return not (self == other)

    def __repr__(self):
        if self.x is None:
            return 'Point(infinity)'
        elif isinstance(self.x, FieldElement):
            return 'Point({},{})_{}_{} FieldElement({})'.format(
                self.x.num, self.y.num, self.a.num, self.b.num, self.x.prime)
        else:
            return 'Point({},{})_{}_{}'.format(self.x, self.y, self.a, self.b)

    def __add__(self, other):
        if self.a != other.a or self.b != other.b:
            raise TypeError('Points {}, {} are not on the same curve'.format(self, other))
        # Case 0.0: self is the point at infinity, return other
        if self.x is None:
            return other
        # Case 0.1: other is the point at infinity, return self
        if other.x is None:
            return self

        # Case 1: self.x == other.x, self.y != other.y
        # Result is point at infinity
        if self.x == other.x and self.y != other.y:
            return self.__class__(None, None, self.a, self.b)

        # Case 2: self.x ≠ other.x
        # Formula (x3,y3)==(x1,y1)+(x2,y2)
        # s=(y2-y1)/(x2-x1)
        # x3=s**2-x1-x2
        # y3=s*(x1-x3)-y1
        if self.x != other.x:
            s = (other.y - self.y) / (other.x - self.x)
            x = s**2 - self.x - other.x
            y = s * (self.x - x) - self.y
            return self.__class__(x, y, self.a, self.b)

        # Case 4: if we are tangent to the vertical line,
        # we return the point at infinity
        # note instead of figuring out what 0 is for each type
        # we just use 0 * self.x
        if self == other and self.y == 0 * self.x:
            return self.__class__(None, None, self.a, self.b)

        # Case 3: self == other
        # Formula (x3,y3)=(x1,y1)+(x1,y1)
        # s=(3*x1**2+a)/(2*y1)
        # x3=s**2-2*x1
        # y3=s*(x1-x3)-y1
        if self == other:
            s = (3 * self.x**2 + self.a) / (2 * self.y)
            x = s**2 - 2 * self.x
            y = s * (self.x - x) - self.y
            return self.__class__(x, y, self.a, self.b)

    def __rmul__(self, coefficient):
        coef = coefficient
        current = self
        result = self.__class__(None, None, self.a, self.b)
        while coef:
            if coef & 1:
                result += current
            current += current
            coef >>= 1
        return result


class S256Field(FieldElement):

    def __init__(self, num, prime=None):
        super().__init__(num=num, prime=P)

    def __repr__(self):
        return '{:x}'.format(self.num).zfill(64)

    def sqrt(self):
        return self**((P + 1) // 4)


class S256Point(Point):

    def __init__(self, x, y, a=None, b=None):
        a, b = S256Field(A), S256Field(B)
        if type(x) == int:
            super().__init__(x=S256Field(x), y=S256Field(y), a=a, b=b)
        else:
            super().__init__(x=x, y=y, a=a, b=b)

    def __repr__(self):
        if self.x is None:
            return 'S256Point(infinity)'
        else:
            return 'S256Point({}, {})'.format(self.x, self.y)

    def __rmul__(self, coefficient):
        coef = coefficient % N
        return super().__rmul__(coef)

    def verify(self, z, sig):
        # By Fermat's Little Theorem, 1/s = pow(s, N-2, N)
        s_inv = pow(sig.s, N - 2, N)
        # u = z / s
        u = z * s_inv % N
        # v = r / s
        v = sig.r * s_inv % N
        # u*G + v*P should have as the x coordinate, r
        total = u * G + v * self
        return total.x.num == sig.r

    def sec(self, compressed=True):
        '''returns the binary version of the SEC format'''
        # if compressed, starts with b'\x02' if self.y.num is even, b'\x03' if self.y is odd
        # then self.x.num
        # remember, you have to convert self.x.num/self.y.num to binary (some_integer.to_bytes(32, 'big'))
        if compressed:
            if self.y.num % 2 == 0:
                return b'\x02' + self.x.num.to_bytes(32, 'big')
            else:
                return b'\x03' + self.x.num.to_bytes(32, 'big')
        else:
            # if non-compressed, starts with b'\x04' followod by self.x and then self.y
            return b'\x04' + self.x.num.to_bytes(32, 'big') + \
                self.y.num.to_bytes(32, 'big')

    def hash160(self, compressed=True):
        return hash160(self.sec(compressed))

    def address(self, compressed=True, testnet=False):
        '''Returns the address string'''
        h160 = self.hash160(compressed)
        if testnet:
            prefix = b'\x6f'
        else:
            prefix = b'\x00'
        return encode_base58_checksum(prefix + h160)

    @classmethod
    def parse(self, sec_bin):
        '''returns a Point object from a SEC binary (not hex)'''
        if sec_bin[0] == 4:
            x = int.from_bytes(sec_bin[1:33], 'big')
            y = int.from_bytes(sec_bin[33:65], 'big')
            return S256Point(x=x, y=y)
        is_even = sec_bin[0] == 2
        x = S256Field(int.from_bytes(sec_bin[1:], 'big'))
        # right side of the equation y^2 = x^3 + 7
        alpha = x**3 + S256Field(B)
        # solve for left side
        beta = alpha.sqrt()
        if beta.num % 2 == 0:
            even_beta = beta
            odd_beta = S256Field(P - beta.num)
        else:
            even_beta = S256Field(P - beta.num)
            odd_beta = beta
        if is_even:
            return S256Point(x, even_beta)
        else:
            return S256Point(x, odd_beta)


G = S256Point(
    0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798,
    0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8
)


class Signature:

    def __init__(self, r, s):
        self.r = r
        self.s = s

    def __repr__(self):
        return 'Signature({:x},{:x})'.format(self.r, self.s)

    def der(self):
        rbin = self.r.to_bytes(32, byteorder='big')
        # remove all null bytes at the beginning
        rbin = rbin.lstrip(b'\x00')
        # if rbin has a high bit, add a \x00
        if rbin[0] & 0x80:
            rbin = b'\x00' + rbin
        result = bytes([2, len(rbin)]) + rbin  # <1>
        sbin = self.s.to_bytes(32, byteorder='big')
        # remove all null bytes at the beginning
        sbin = sbin.lstrip(b'\x00')
        # if sbin has a high bit, add a \x00
        if sbin[0] & 0x80:
            sbin = b'\x00' + sbin
        result += bytes([2, len(sbin)]) + sbin
        return bytes([0x30, len(result)]) + result

    @classmethod
    def parse(cls, signature_bin):
        s = BytesIO(signature_bin)
        compound = s.read(1)[0]
        if compound != 0x30:
            raise SyntaxError("Bad Signature")
        length = s.read(1)[0]
        if length + 2 != len(signature_bin):
            raise SyntaxError("Bad Signature Length")
        marker = s.read(1)[0]
        if marker != 0x02:
            raise SyntaxError("Bad Signature")
        rlength = s.read(1)[0]
        r = int.from_bytes(s.read(rlength), 'big')
        marker = s.read(1)[0]
        if marker != 0x02:
            raise SyntaxError("Bad Signature")
        slength = s.read(1)[0]
        s = int.from_bytes(s.read(slength), 'big')
        if len(signature_bin) != 6 + rlength + slength:
            raise SyntaxError("Signature too long")
        return cls(r, s)


class PrivateKey:

    def __init__(self, secret):
        self.secret = secret
        self.point = secret * G

    def hex(self):
        return '{:x}'.format(self.secret).zfill(64)

    def sign(self, z):
        k = self.deterministic_k(z)
        # r is the x coordinate of the resulting point k*G
        r = (k * G).x.num
        # remember 1/k = pow(k, N-2, N)
        k_inv = pow(k, N - 2, N)
        # s = (z+r*secret) / k
        s = (z + r * self.secret) * k_inv % N
        if s > N / 2:
            s = N - s
        # return an instance of Signature:
        # Signature(r, s)
        return Signature(r, s)

    def deterministic_k(self, z):
        k = b'\x00' * 32
        v = b'\x01' * 32
        if z > N:
            z -= N
        z_bytes = z.to_bytes(32, 'big')
        secret_bytes = self.secret.to_bytes(32, 'big')
        s256 = hashlib.sha256
        k = hmac.new(k, v + b'\x00' + secret_bytes + z_bytes, s256).digest()
        v = hmac.new(k, v, s256).digest()
        k = hmac.new(k, v + b'\x01' + secret_bytes + z_bytes, s256).digest()
        v = hmac.new(k, v, s256).digest()
        while True:
            v = hmac.new(k, v, s256).digest()
            candidate = int.from_bytes(v, 'big')
            if candidate >= 1 and candidate < N:
                return candidate
            k = hmac.new(k, v + b'\x00', s256).digest()
            v = hmac.new(k, v, s256).digest()

    def wif(self, compressed=True, testnet=False):
        # convert the secret from integer to a 32-bytes in big endian using num.to_bytes(32, 'big')
        secret_bytes = self.secret.to_bytes(32, 'big')
        # prepend b'\xef' on testnet, b'\x80' on mainnet
        if testnet:
            prefix = b'\xef'
        else:
            prefix = b'\x80'
        # append b'\x01' if compressed
        if compressed:
            suffix = b'\x01'
        else:
            suffix = b''
        # encode_base58_checksum the whole thing
        return encode_base58_checksum(prefix + secret_bytes + suffix)


#!/usr/bin/env python3

import json
import unicodedata
import os

from hdwallet.utils import (
    get_bytes, generate_mnemonic, generate_entropy, is_entropy, get_mnemonic_strength,
    get_entropy_strength, is_mnemonic, get_mnemonic_language, entropy_to_mnemonic, mnemonic_to_entropy,
    is_root_xprivate_key, is_root_xpublic_key, generate_passphrase
)

# Test Values
base_path: str = os.path.dirname(__file__)
file_path: str = os.path.abspath(os.path.join(base_path, "values.json"))
values = open(file_path, "r", encoding="utf-8")
_: dict = json.loads(values.read())
values.close()


def test_utils():

    assert isinstance(get_bytes(string=b"meherett"), bytes)

    assert is_root_xprivate_key(
        xprivate_key=_["bitcoin"]["mainnet"]["root_xprivate_key"],
        symbol=_["bitcoin"]["mainnet"]["symbol"]
    )
    assert not is_root_xprivate_key(
        xprivate_key=_["bitcoin"]["mainnet"]["xprivate_key"],
        symbol=_["bitcoin"]["mainnet"]["symbol"]
    )

    assert is_root_xpublic_key(
        xpublic_key=_["bitcoin"]["mainnet"]["root_xpublic_key"],
        symbol=_["bitcoin"]["mainnet"]["symbol"]
    )
    assert not is_root_xpublic_key(
        xpublic_key=_["bitcoin"]["mainnet"]["xpublic_key"],
        symbol=_["bitcoin"]["mainnet"]["symbol"]
    )

    assert len(generate_passphrase(length=19999)) == 19999


def test_utils_entropy():

    assert len(generate_entropy(strength=128)) == 32
    assert len(generate_entropy(strength=160)) == 40
    assert len(generate_entropy(strength=192)) == 48
    assert len(generate_entropy(strength=224)) == 56
    assert len(generate_entropy(strength=256)) == 64

    for entropy in _["utils"]["entropy"]:

        assert len(entropy["entropy"]) == entropy["length"]
        assert get_entropy_strength(entropy["entropy"]) == entropy["strength"]
        assert is_entropy(entropy["entropy"])

    assert mnemonic_to_entropy(
        mnemonic=_["bitcoin"]["mainnet"]["mnemonic"], language=_["bitcoin"]["mainnet"]["language"]
    ) == _["bitcoin"]["mainnet"]["entropy"]


def test_utils_mnemonic():

    assert len(generate_mnemonic(strength=128).split(" ")) == 12
    assert len(generate_mnemonic(strength=160).split(" ")) == 15
    assert len(generate_mnemonic(strength=192).split(" ")) == 18
    assert len(generate_mnemonic(strength=224).split(" ")) == 21
    assert len(generate_mnemonic(strength=256).split(" ")) == 24

    for mnemonic in _["utils"]["mnemonic"]:

        assert len(unicodedata.normalize("NFKD", mnemonic["mnemonic"]).split(" ")) == mnemonic["words"]
        assert get_mnemonic_strength(mnemonic["mnemonic"]) == mnemonic["strength"]
        assert is_mnemonic(mnemonic["mnemonic"])
        if mnemonic["language"] == "english":
            assert not is_mnemonic(mnemonic["mnemonic"], "korean")
        assert is_mnemonic(mnemonic["mnemonic"], mnemonic["language"])
        assert get_mnemonic_language(mnemonic["mnemonic"]) == mnemonic["language"]

    assert entropy_to_mnemonic(
        entropy=_["bitcoin"]["mainnet"]["entropy"], language=_["bitcoin"]["mainnet"]["language"]
    ) == _["bitcoin"]["mainnet"]["mnemonic"]



#!/usr/bin/env python3

import json
import os

from hdwallet.cli.__main__ import main as cli_main

# Test Values
base_path = os.path.dirname(__file__)
file_path = os.path.abspath(os.path.join(base_path, "..", "..", "values.json"))
values = open(file_path, "r", encoding="utf-8")
_ = json.loads(values.read())
values.close()


def test_hdwallet(cli_tester):

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--entropy", _["bitcoin"]["mainnet"]["entropy"],
            "--passphrase", _["bitcoin"]["mainnet"]["passphrase"],
            "--language", _["bitcoin"]["mainnet"]["language"]
        ]
    )

    dumps: dict = _["bitcoin"]["mainnet"]

    del dumps["root_xprivate_key_hex"]
    del dumps["root_xpublic_key_hex"]
    del dumps["xprivate_key_hex"]
    del dumps["xpublic_key_hex"]

    assert hdwallet.exit_code == 0
    assert hdwallet.output == str(json.dumps(dumps, indent=4, ensure_ascii=False)) + "\n"

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--mnemonic", _["bitcoin"]["mainnet"]["mnemonic"],
            "--passphrase", _["bitcoin"]["mainnet"]["passphrase"],
            "--language", _["bitcoin"]["mainnet"]["language"]
        ]
    )

    assert hdwallet.exit_code == 0
    assert hdwallet.output == str(json.dumps(dumps, indent=4, ensure_ascii=False)) + "\n"

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--seed", _["bitcoin"]["mainnet"]["seed"]
        ]
    )

    dumps["strength"] = None
    dumps["entropy"] = None
    dumps["mnemonic"] = None
    dumps["language"] = None

    assert hdwallet.exit_code == 0
    assert hdwallet.output == str(json.dumps(dumps, indent=4, ensure_ascii=False)) + "\n"

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--xprivate-key", _["bitcoin"]["mainnet"]["root_xprivate_key"]
        ]
    )

    dumps["strength"] = None
    dumps["entropy"] = None
    dumps["mnemonic"] = None
    dumps["language"] = None
    dumps["seed"] = None

    assert hdwallet.exit_code == 0
    assert hdwallet.output == str(json.dumps(dumps, indent=4, ensure_ascii=False)) + "\n"

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--xpublic-key", _["bitcoin"]["mainnet"]["root_xpublic_key"]
        ]
    )

    assert hdwallet.exit_code == 0
    assert hdwallet.output

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--private-key", _["bitcoin"]["mainnet"]["private_key"]
        ]
    )

    dumps["root_xprivate_key"] = None
    dumps["root_xpublic_key"] = None
    dumps["xprivate_key"] = None
    dumps["xpublic_key"] = None
    dumps["chain_code"] = None
    dumps["path"] = None

    assert hdwallet.exit_code == 0
    assert hdwallet.output == str(json.dumps(dumps, indent=4, ensure_ascii=False)) + "\n"

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--wif", _["bitcoin"]["mainnet"]["wif"]
        ]
    )

    assert hdwallet.exit_code == 0
    assert hdwallet.output == str(json.dumps(dumps, indent=4, ensure_ascii=False)) + "\n"

    hdwallet = cli_tester.invoke(
        cli_main, [
            "generate",
            "--symbol", _["bitcoin"]["mainnet"]["symbol"],
            "--public-key", _["bitcoin"]["mainnet"]["public_key"]
        ]
    )

    dumps["private_key"] = None
    dumps["wif"] = None

    assert hdwallet.exit_code == 0
    assert hdwallet.output == str(json.dumps(dumps, indent=4, ensure_ascii=False)) + "\n"



#!/usr/bin/env python

from tabulate import tabulate

import inspect

from hdwallet import cryptocurrencies
from hdwallet.cli import click


def list_cryptocurrencies():

    documents, table, headers = [], [], [
        "Cryptocurrency", "Symbol", "Mainnet", "Testnet", "Segwit", "Coin Type", "Default Path"
    ]

    for name, cryptocurrency in inspect.getmembers(cryptocurrencies):
        if inspect.isclass(cryptocurrency):
            if issubclass(cryptocurrency, cryptocurrencies.Cryptocurrency) \
                    and cryptocurrency != cryptocurrencies.Cryptocurrency:

                if cryptocurrency.NETWORK == "mainnet":
                    document: dict = {
                        "name": cryptocurrency.NAME,
                        "symbol": cryptocurrency.SYMBOL,
                        "source_code": cryptocurrency.SOURCE_CODE,
                        "mainnet": "Yes" if cryptocurrency.NETWORK == "mainnet" else "No",
                        "testnet": "Yes" if cryptocurrency.NETWORK == "testnet" else "No",
                        "segwit": "Yes" if cryptocurrency.SEGWIT_ADDRESS.HRP else "No",
                        "coin_type": cryptocurrency.COIN_TYPE.INDEX,
                        "default_path": cryptocurrency.DEFAULT_PATH
                    }
                    documents.append(document)
                elif cryptocurrency.NETWORK == "testnet":
                    for index, document in enumerate(documents):
                        if document["name"] == cryptocurrency.NAME:
                            documents[index]["symbol"] = f"{document['symbol']}, {cryptocurrency.SYMBOL}"
                            documents[index]["testnet"] = "Yes"
                else:
                    raise Exception("Invalid cryptocurrency network type.")

    for document in documents:
        table.append([
            document["name"],
            document["symbol"],
            document["mainnet"],
            document["testnet"],
            document["segwit"],
            document["coin_type"],
            document["default_path"]
        ])

    click.echo(tabulate(table, headers, tablefmt="github"))



#!/usr/bin/env python3

from hdwallet.cli.__main__ import main as cli_main
from hdwallet import __version__


def test_hdwallet_cli(cli_tester):

    assert cli_tester.invoke(cli_main).exit_code == 0

    assert cli_tester.invoke(cli_main, ["generate"]).exit_code == 0

    version = cli_tester.invoke(cli_main, ["--version"])
    assert version.exit_code == 0
    assert version.output == "%s\n" % __version__



#!/usr/bin/env python

from tabulate import tabulate

from hdwallet.cli import click


def list_languages():

    click.echo(tabulate(
        [
            ["Chinese Simplified"],
            ["Chinese Traditional"],
            ["English"],
            ["French"],
            ["Italian"],
            ["Japanese"],
            ["Korean"],
            ["Spanish"],
        ],
        [
            "Language"
        ],
        tablefmt="github"
    ))



#!/usr/bin/env python3

import json
import os

from hdwallet import HDWallet

# Test Values
base_path: str = os.path.dirname(__file__)
file_path: str = os.path.abspath(os.path.join(base_path, "../values.json"))
values = open(file_path, "r", encoding="utf-8")
_: dict = json.loads(values.read())
values.close()


def test_from_mnemonic():

    hdwallet: HDWallet = HDWallet(
        symbol=_["bitcoin"]["mainnet"]["symbol"]
    )
    
    hdwallet.from_mnemonic(
        mnemonic=_["bitcoin"]["mainnet"]["mnemonic"],
        passphrase=_["bitcoin"]["mainnet"]["passphrase"],
        language=_["bitcoin"]["mainnet"]["language"]
    )

    hdwallet.from_path(
        path=_["bitcoin"]["mainnet"]["path"]
    )

    assert hdwallet.cryptocurrency() == _["bitcoin"]["mainnet"]["cryptocurrency"]
    assert hdwallet.symbol() == _["bitcoin"]["mainnet"]["symbol"]
    assert hdwallet.network() == _["bitcoin"]["mainnet"]["network"]
    assert hdwallet.strength() == _["bitcoin"]["testnet"]["strength"]
    assert hdwallet.entropy() == _["bitcoin"]["mainnet"]["entropy"]
    assert hdwallet.mnemonic() == _["bitcoin"]["mainnet"]["mnemonic"]
    assert hdwallet.language() == _["bitcoin"]["mainnet"]["language"]
    assert hdwallet.passphrase() == _["bitcoin"]["mainnet"]["passphrase"]
    assert hdwallet.seed() == _["bitcoin"]["mainnet"]["seed"]
    assert hdwallet.root_xprivate_key(encoded=False) == _["bitcoin"]["mainnet"]["root_xprivate_key_hex"]
    assert hdwallet.root_xprivate_key() == _["bitcoin"]["mainnet"]["root_xprivate_key"]
    assert hdwallet.root_xpublic_key(encoded=False) == _["bitcoin"]["mainnet"]["root_xpublic_key_hex"]
    assert hdwallet.root_xpublic_key() == _["bitcoin"]["mainnet"]["root_xpublic_key"]
    assert hdwallet.xprivate_key(encoded=False) == _["bitcoin"]["mainnet"]["xprivate_key_hex"]
    assert hdwallet.xprivate_key() == _["bitcoin"]["mainnet"]["xprivate_key"]
    assert hdwallet.xpublic_key(encoded=False) == _["bitcoin"]["mainnet"]["xpublic_key_hex"]
    assert hdwallet.xpublic_key() == _["bitcoin"]["mainnet"]["xpublic_key"]
    assert hdwallet.uncompressed() == _["bitcoin"]["mainnet"]["uncompressed"]
    assert hdwallet.compressed() == _["bitcoin"]["mainnet"]["compressed"]
    assert hdwallet.chain_code() == _["bitcoin"]["mainnet"]["chain_code"]
    assert hdwallet.private_key() == _["bitcoin"]["mainnet"]["private_key"]
    assert hdwallet.public_key() == _["bitcoin"]["mainnet"]["public_key"]
    assert hdwallet.wif() == _["bitcoin"]["mainnet"]["wif"]
    assert hdwallet.finger_print() == _["bitcoin"]["mainnet"]["finger_print"]
    assert hdwallet.semantic() == _["bitcoin"]["mainnet"]["semantic"]
    assert hdwallet.path() == _["bitcoin"]["mainnet"]["path"]
    assert hdwallet.hash() == _["bitcoin"]["mainnet"]["hash"]
    assert hdwallet.p2pkh_address() == _["bitcoin"]["mainnet"]["addresses"]["p2pkh"]
    assert hdwallet.p2sh_address() == _["bitcoin"]["mainnet"]["addresses"]["p2sh"]
    assert hdwallet.p2wpkh_address() == _["bitcoin"]["mainnet"]["addresses"]["p2wpkh"]
    assert hdwallet.p2wpkh_in_p2sh_address() == _["bitcoin"]["mainnet"]["addresses"]["p2wpkh_in_p2sh"]
    assert hdwallet.p2wsh_address() == _["bitcoin"]["mainnet"]["addresses"]["p2wsh"]
    assert hdwallet.p2wsh_in_p2sh_address() == _["bitcoin"]["mainnet"]["addresses"]["p2wsh_in_p2sh"]

    assert isinstance(hdwallet.dumps(), dict)

    dumps: dict = _["bitcoin"]["mainnet"]

    del dumps["root_xprivate_key_hex"]
    del dumps["root_xpublic_key_hex"]
    del dumps["xprivate_key_hex"]
    del dumps["xpublic_key_hex"]

    assert hdwallet.dumps() == dumps



#!/usr/bin/env python3

from binascii import (
    hexlify, unhexlify
)

import pytest

from hdwallet.libs.base58 import (
    check_encode, check_decode, decode, encode, string_to_int
)


RAW: str = "0488ade405d5bc481680000000b5b40efbdef2e0a2e8ee6fa9d371f951f93e3b1e3c1b50fb31f9ea36cfc7c88100b9d3f3" \
           "12b38951361597219c8979e177d99349fd0213bac55508c8469d02064f"

ENCODED_RAW: str = "xpub6GzA69ujvGDmXQDwJnCnyDBr5FxBFri57tFBzD1Q3HTUkRRS35eZND12MPL1d7iXT15tLwbKmvxhJDFtExNc7SU" \
                   "efgYGNE7xCQan6Cmhp8f"


def test_base58():

    assert check_encode(raw=unhexlify(RAW)) == "xprvA3zogeNr5tfUJv9UCkfnc5F7XE7grPzDkfKbBpbnUwvVsd6HVYLJp" \
                                               "QgYWCBjXwZw3GdtSVUQ7BDEgmi6MHt4MatV9gDtWuJuGYhYL7nj6aa"

    assert hexlify(check_decode(enc=ENCODED_RAW)).decode() == \
        "0488b21e05d5bc481680000000b5b40efbdef2e0a2e8ee6fa9d371f951f93e3b1e3c1b50fb31f9ea36cfc7c88100b9d3" \
        "f312b38951361597219c8979e177d99349fd0213bac55508c8469d02064f"

    with pytest.raises(TypeError, match="string argument without an encoding"):
        assert string_to_int(str("meherett"))


    assert decode("111233QC4") == b'\x00\x00\x00(\x7f\xb4\xcd'


    assert encode(decode("111233QC4")) == "111233QC4"



#!/usr/bin/env python

from typing import Optional

from hdwallet import HDWallet
from hdwallet.derivations import (
    Derivation, BIP32Derivation
)
from hdwallet.cryptocurrencies import (
    Cryptocurrency, get_cryptocurrency
)
from hdwallet.utils import generate_mnemonic
from hdwallet.cli import (
    click, sys
)


def generate_addresses(
    symbol: str,
    strength: int,
    entropy: Optional[str],
    mnemonic: Optional[str],
    language: Optional[str],
    passphrase: Optional[str],
    seed: Optional[str],
    xprivate_key: Optional[str],
    xpublic_key: Optional[str],
    strict: bool,
    account: int,
    change: bool,
    path: Optional[str],
    semantic: str,
    start_index: int,
    end_index: int,
    hardened: bool,
    show: str
):
    try:
        hdwallet: HDWallet = HDWallet(
            symbol=symbol, semantic=semantic
        )
        if entropy:
            hdwallet.from_entropy(
                entropy=entropy, language=language, passphrase=passphrase
            )
        elif mnemonic:
            hdwallet.from_mnemonic(
                mnemonic=mnemonic, language=language, passphrase=passphrase
            )
        elif seed:
            hdwallet.from_seed(
                seed=seed
            )
        elif xprivate_key:
            hdwallet.from_xprivate_key(
                xprivate_key=xprivate_key, strict=strict
            )
        elif xpublic_key:
            hdwallet.from_xpublic_key(
                xpublic_key=xpublic_key, strict=strict
            )
        else:
            mnemonic = generate_mnemonic(language=language, strength=strength)
            hdwallet.from_mnemonic(
                mnemonic=mnemonic, language=language, passphrase=passphrase
            )

        for index in range(start_index, end_index):
            if path:
                derivation: Derivation = Derivation(path=path)
                derivation.from_index(index=index, hardened=hardened)
                hdwallet.from_path(path=derivation)
            else:
                cryptocurrency: Cryptocurrency = get_cryptocurrency(symbol=symbol)
                bip32_derivation: BIP32Derivation = BIP32Derivation(
                    purpose=(
                        44, False if xpublic_key else True
                    ),
                    coin_type=(
                        cryptocurrency.COIN_TYPE.INDEX,
                        False if xpublic_key else cryptocurrency.COIN_TYPE.HARDENED
                    ),
                    account=(
                        account, False if xpublic_key else True
                    ),
                    change=change,
                    address=index
                )
                hdwallet.from_path(path=bip32_derivation)

            rows: str = ""
            dumps = hdwallet.dumps()
            for i, key in enumerate([keys.split(":") for keys in show.split(",")]):
                rows += (
                    f"{dumps[key[0]][key[1]] if len(key) == 2 else dumps[key[0]]}"
                    if i == 0 else
                    f" {dumps[key[0]][key[1]] if len(key) == 2 else dumps[key[0]]}"
                )
            click.echo(rows)

            hdwallet.clean_derivation()

    except TimeoutError as exception:
        click.echo(click.style(f"Error: {str(exception)}"), err=True)
        sys.exit()



#!/usr/bin/env python3

import pytest

from hdwallet.cryptocurrencies import BitcoinMainnet
from hdwallet.derivations import (
    Derivation, BIP32Derivation, BIP44Derivation, BIP49Derivation, BIP84Derivation, BIP141Derivation
)
from hdwallet.exceptions import SemanticError


def test_derivations():

    assert str(
        Derivation(path="m/44'/0'/0'/0/0", semantic="p2pkh")
    ) == "m/44'/0'/0'/0/0"
    assert str(
        Derivation(semantic="p2pkh").from_path(
            path="m/44'/0'/0'/0/0"
        )
    ) == "m/44'/0'/0'/0/0"
    assert str(
        Derivation(semantic="p2pkh").from_index(
            index=44, hardened=True
        ).from_index(
            index=0, hardened=True
        ).from_index(
            index=0, hardened=True
        ).from_index(
            index=0, hardened=False
        ).from_index(
            index=0, hardened=False
        )
    ) == "m/44'/0'/0'/0/0"
    assert str(
        Derivation(path="m/44'/0'/0'/0/0", semantic="p2pkh").clean_derivation()
    ) == "\0\0\0\0"

    assert str(
        BIP32Derivation(purpose=44, coin_type=0, account=0, change=False, address=0)
    ) == "m/44'/0'/0'/0/0"
    assert str(
        BIP32Derivation(purpose=(44, False), coin_type=(0, False), account=0, change=False, address=0)
    ) == "m/44/0/0'/0/0"

    bip32_derivation: BIP32Derivation = BIP32Derivation().from_purpose(
        purpose=44, hardened=True
    ).from_coin_type(
        coin_type=0, hardened=True
    ).from_account(
        account=0, hardened=True
    ).from_change(
        change=False
    ).from_address(
        address=0, hardened=False
    )
    assert str(bip32_derivation) == "m/44'/0'/0'/0/0"
    assert bip32_derivation.purpose() == "44'"
    assert bip32_derivation.coin_type() == "0'"
    assert bip32_derivation.account() == "0'"
    assert bip32_derivation.change() == 0
    assert bip32_derivation.address() == "0"
    assert bip32_derivation.SEMANTIC == "p2pkh"

    bip32_derivation: BIP32Derivation = BIP32Derivation(
        purpose=(44, True), coin_type=(0, True), account=(0, True), change=False, address=(0, False)
    )
    assert str(bip32_derivation) == "m/44'/0'/0'/0/0"
    assert bip32_derivation.purpose() == "44'"
    assert bip32_derivation.coin_type() == "0'"
    assert bip32_derivation.account() == "0'"
    assert bip32_derivation.change() == 0
    assert bip32_derivation.address() == "0"
    assert bip32_derivation.SEMANTIC == "p2pkh"

    bip44_derivation: BIP44Derivation = BIP44Derivation(
        cryptocurrency=BitcoinMainnet
    ).from_account(
        account=0, hardened=True
    ).from_change(
        change=False
    ).from_address(
        address=0, hardened=False
    )
    assert str(bip44_derivation) == "m/44'/0'/0'/0/0"
    assert bip44_derivation.purpose() == "44'"
    assert bip44_derivation.coin_type() == "0'"
    assert bip44_derivation.account() == "0'"
    assert bip44_derivation.change() == 0
    assert bip44_derivation.address() == "0"
    assert bip44_derivation.SEMANTIC == "p2pkh"

    bip44_derivation: BIP44Derivation = BIP44Derivation(
        cryptocurrency=BitcoinMainnet, account=(0, True), change=False, address=(0, False)
    )
    assert str(bip44_derivation) == "m/44'/0'/0'/0/0"
    assert bip44_derivation.purpose() == "44'"
    assert bip44_derivation.coin_type() == "0'"
    assert bip44_derivation.account() == "0'"
    assert bip44_derivation.change() == 0
    assert bip44_derivation.address() == "0"
    assert bip44_derivation.SEMANTIC == "p2pkh"

    bip49_derivation: BIP49Derivation = BIP49Derivation(
        cryptocurrency=BitcoinMainnet
    ).from_account(
        account=0, hardened=True
    ).from_change(
        change=False
    ).from_address(
        address=0, hardened=False
    )
    assert str(bip49_derivation) == "m/49'/0'/0'/0/0"
    assert bip49_derivation.purpose() == "49'"
    assert bip49_derivation.coin_type() == "0'"
    assert bip49_derivation.account() == "0'"
    assert bip49_derivation.change() == 0
    assert bip49_derivation.address() == "0"
    assert bip49_derivation.SEMANTIC == "p2wpkh_in_p2sh"

    bip49_derivation: BIP49Derivation = BIP49Derivation(
        cryptocurrency=BitcoinMainnet, account=(0, True), change=False, address=(0, False)
    )
    assert str(bip49_derivation) == "m/49'/0'/0'/0/0"
    assert bip49_derivation.purpose() == "49'"
    assert bip49_derivation.coin_type() == "0'"
    assert bip49_derivation.account() == "0'"
    assert bip49_derivation.change() == 0
    assert bip49_derivation.address() == "0"
    assert bip49_derivation.SEMANTIC == "p2wpkh_in_p2sh"

    bip84_derivation: BIP84Derivation = BIP84Derivation(
        cryptocurrency=BitcoinMainnet
    ).from_account(
        account=0, hardened=True
    ).from_change(
        change=False
    ).from_address(
        address=0, hardened=False
    )
    assert str(bip84_derivation) == "m/84'/0'/0'/0/0"
    assert bip84_derivation.purpose() == "84'"
    assert bip84_derivation.coin_type() == "0'"
    assert bip84_derivation.account() == "0'"
    assert bip84_derivation.change() == 0
    assert bip84_derivation.address() == "0"
    assert bip84_derivation.SEMANTIC == "p2wpkh"

    bip84_derivation: BIP84Derivation = BIP84Derivation(
        cryptocurrency=BitcoinMainnet, account=(0, True), change=False, address=(0, False)
    )
    assert str(bip84_derivation) == "m/84'/0'/0'/0/0"
    assert bip84_derivation.purpose() == "84'"
    assert bip84_derivation.coin_type() == "0'"
    assert bip84_derivation.account() == "0'"
    assert bip84_derivation.change() == 0
    assert bip84_derivation.address() == "0"
    assert bip84_derivation.SEMANTIC == "p2wpkh"

    bip141_derivation: BIP141Derivation = BIP141Derivation(
        cryptocurrency=BitcoinMainnet, semantic="p2wsh_in_p2sh"
    )
    assert str(bip141_derivation) == "m/44'/0'/0'/0/0"
    assert bip141_derivation.SEMANTIC == "p2wsh_in_p2sh"

    with pytest.raises(SemanticError, match="Wrong extended semantic, choose only the following options 'p2wpkh', 'p2wpkh_in_p2sh', 'p2wsh' or 'p2wsh_in_p2sh' semantics."):
        BIP141Derivation(
            cryptocurrency=BitcoinMainnet, semantic="p2pkh"
        )
